@@ -1,8 +1,9 @@
-use crate::ModularInteger;
+use crate::{ModularInteger, udouble};
 use num_integer::Integer;
 use num_traits::Pow;
-use std::borrow::Borrow;
-use std::ops::{Add, Mul, Neg, Sub};
+use core::ops::{Add, Mul, Neg, Sub};
+
+#[cfg(std)]
 use std::rc::Rc;
 
 /// Operations of a integer represented in Montgomery form. This data type can
@@ -59,7 +60,7 @@ const BINVERT_TABLE: [u8; 128] = [
     0x21, 0xCB, 0xED, 0xD7, 0x59, 0xC3, 0xE5, 0x0F, 0x11, 0x3B, 0x5D, 0xC7, 0x49, 0x33, 0x55, 0xFF,
 ];
 
-macro_rules! impl_uprim_montgomery {
+macro_rules! impl_uprim_montgomery_core {
     () => {
         #[inline]
         fn transform(target: Self, m: &Self) -> Self {
@@ -69,8 +70,8 @@ macro_rules! impl_uprim_montgomery {
             (((target as Self::Double) << Self::BITS) % (*m as Self::Double)) as _
         }
 
+        // REDC algorithm
         fn reduce(monty: Self::Double, m: &Self, minv: &Self::Inv) -> Self {
-            // REDC algorithm
             debug_assert!(monty < ((*m as Self::Double) << Self::BITS));
 
             let tm = (monty as Self).wrapping_mul(*minv);
@@ -87,6 +88,14 @@ macro_rules! impl_uprim_montgomery {
             }
         }
 
+        #[inline]
+        fn mul(lhs: &Self, rhs: &Self, m: &Self, minv: &Self::Inv) -> Self {
+            Montgomery::reduce((*lhs as Self::Double) * (*rhs as Self::Double), m, minv)
+        }
+    }
+}
+macro_rules! impl_uprim_montgomery {
+    () => {
         #[inline]
         fn add(lhs: &Self, rhs: &Self, m: &Self) -> Self {
             let (sum, overflow) = lhs.overflowing_add(*rhs);
@@ -115,11 +124,6 @@ macro_rules! impl_uprim_montgomery {
             }
         }
 
-        #[inline]
-        fn mul(lhs: &Self, rhs: &Self, m: &Self, minv: &Self::Inv) -> Self {
-            Montgomery::reduce((*lhs as Self::Double) * (*rhs as Self::Double), m, minv)
-        }
-
         fn pow(base: &Self, exp: &Self, m: &Self, minv: &Self::Inv) -> Self {
             match *exp {
                 1 => *base,
@@ -127,7 +131,7 @@ macro_rules! impl_uprim_montgomery {
                 e => {
                     let mut multi = *base;
                     let mut exp = e;
-                    let mut result = Montgomery::transform(1, m);
+                    let mut result = Montgomery::transform(1, m); // TODO: subtract exp by 1 and use base?
                     while exp > 0 {
                         if exp & 1 != 0 {
                             result = Montgomery::mul(&result, &multi, m, minv);
@@ -146,57 +150,103 @@ impl Montgomery for u8 {
     type Inv = u8;
     type Double = u16;
 
+    impl_uprim_montgomery_core!();
+    impl_uprim_montgomery!();
+
     fn neginv(m: &Self) -> Self {
         BINVERT_TABLE[((m >> 1) & 0x7F) as usize].wrapping_neg()
     }
-
-    impl_uprim_montgomery!();
 }
 
 impl Montgomery for u16 {
     type Inv = u16;
     type Double = u32;
 
+    impl_uprim_montgomery_core!();
+    impl_uprim_montgomery!();
+
     fn neginv(m: &Self) -> Self {
         let i = BINVERT_TABLE[((m >> 1) & 0x7F) as usize] as u16;
-        // Newton-Rhapson iteration
-        // See: https://arxiv.org/abs/1303.0328
+        // Newton-Rhapson iteration (hensel lifting), see https://arxiv.org/abs/1303.0328
         i.wrapping_mul(*m).wrapping_sub(2).wrapping_mul(i)
     }
-
-    impl_uprim_montgomery!();
 }
 
 impl Montgomery for u32 {
     type Inv = u32;
     type Double = u64;
 
+    impl_uprim_montgomery_core!();
+    impl_uprim_montgomery!();
+
     fn neginv(m: &Self) -> Self {
         let i = BINVERT_TABLE[((m >> 1) & 0x7F) as usize] as u32;
         let i = 2u32.wrapping_sub(i.wrapping_mul(*m)).wrapping_mul(i);
         i.wrapping_mul(*m).wrapping_sub(2).wrapping_mul(i)
     }
-
-    impl_uprim_montgomery!();
 }
 
 impl Montgomery for u64 {
     type Inv = u64;
     type Double = u128;
 
+    impl_uprim_montgomery_core!();
+    impl_uprim_montgomery!();
+
     fn neginv(m: &Self) -> Self {
         let i = BINVERT_TABLE[((m >> 1) & 0x7F) as usize] as u64;
         let i = 2u64.wrapping_sub(i.wrapping_mul(*m)).wrapping_mul(i);
         let i = 2u64.wrapping_sub(i.wrapping_mul(*m)).wrapping_mul(i);
         i.wrapping_mul(*m).wrapping_sub(2).wrapping_mul(i)
     }
+}
+
+impl Montgomery for u128 {
+    type Inv = u128;
+    type Double = udouble;
+
+    fn neginv(m: &Self) -> Self {
+        let i = BINVERT_TABLE[((m >> 1) & 0x7F) as usize] as u128;
+        let i = 2u128.wrapping_sub(i.wrapping_mul(*m)).wrapping_mul(i);
+        let i = 2u128.wrapping_sub(i.wrapping_mul(*m)).wrapping_mul(i);
+        let i = 2u128.wrapping_sub(i.wrapping_mul(*m)).wrapping_mul(i);
+        i.wrapping_mul(*m).wrapping_sub(2).wrapping_mul(i)
+    }
+    
+    #[inline]
+    fn transform(target: Self, m: &Self) -> Self {
+        if target == 0 {
+            return 0;
+        }
+        let r = udouble { hi: target, lo: 0 } % udouble::from(*m);
+        r.lo
+    }
+
+    // REDC algorithm
+    fn reduce(monty: Self::Double, m: &Self, minv: &Self::Inv) -> Self {
+        debug_assert!(monty < udouble { hi: *m, lo: 0 });
+
+        let tm = monty.lo.wrapping_mul(*minv);
+        let (t, overflow) = monty.overflowing_add(udouble::widening_mul(tm, *m));
+
+        // in case of overflow, we need to add another `R mod m` = `R - m`
+        let t = if overflow { t.hi + m.wrapping_neg() } else { t.hi };
+
+        if &t >= m {
+            return t - m;
+        } else {
+            return t;
+        }
+    }
+
+    #[inline]
+    fn mul(lhs: &Self, rhs: &Self, m: &Self, minv: &Self::Inv) -> Self {
+        Montgomery::reduce(udouble::widening_mul(*lhs, *rhs), m, minv)
+    }
 
     impl_uprim_montgomery!();
 }
 
-// XXX: implement Montgomery for u128 (double type is also u128), which requires efficient implementation of dual word mul_mod.
-// REF: https://github.com/coreutils/coreutils/blob/master/src/factor.c (mulredc2)
-
 /// An integer represented in Montgomery form, it implements [ModularInteger] interface
 /// and it's generally more efficient than the vanilla integer in modular operations.
 #[derive(Debug, Clone, Copy)]
@@ -211,21 +261,13 @@ pub struct MontgomeryInt<T: Integer + Montgomery> {
     mi: T::Inv,
 }
 
-/// A big integer represented in Montgomery form, it implements [ModularInteger] interface
-/// and it's generally more efficient than the vanilla integer in modular operations.
-///
-/// The modulus is stored in heap to prevent frequent copying
-#[derive(Debug, Clone)]
-pub struct MontgomeryBigint<T: Integer + Montgomery> {
-    /// The Montgomery representation of the integer.
-    a: T,
+/// A word-size integer in Montgomery form with fixed modulus
+// TODO: implement after we have const implementation of invm
+#[derive(Debug, Clone, Copy)]
+struct MontgomeryWord<const M: usize> (usize);
 
-    /// The modulus and its negated modular inverse.
-    ///
-    /// It's stored as a pointer to prevent frequent copying. It also allows
-    /// quick checking of the equity of two moduli.
-    minv: Rc<(T, T::Inv)>,
-}
+// XXX: we can also implement MontgomeryMersenne<const M: usize> to support Montgomery form
+// with (Pseudo) Mersenne prime as modulo. REF: https://eprint.iacr.org/2018/1038.pdf
 
 impl<T: Integer + Montgomery> MontgomeryInt<T> {
     #[inline]
@@ -236,17 +278,6 @@ impl<T: Integer + Montgomery> MontgomeryInt<T> {
     }
 }
 
-impl<T: Integer + Montgomery> MontgomeryBigint<T> {
-    #[inline]
-    fn check_modulus_eq(&self, rhs: &Self) {
-        if Rc::ptr_eq(&self.minv, &rhs.minv) {
-            if self.minv.0 != rhs.minv.0 {
-                panic!("The modulus of two operators should be the same!");
-            }
-        }
-    }
-}
-
 impl<T: Integer + Montgomery> MontgomeryInt<T>
 where
     T::Double: From<T>,
@@ -260,22 +291,6 @@ where
     }
 }
 
-impl<T: Integer + Montgomery> MontgomeryBigint<T>
-where
-    T::Double: From<T>,
-{
-    /// Convert n into the modulo ring ℤ/mℤ (i.e. `n % m`)
-    #[inline]
-    pub fn new(n: T, m: T) -> Self {
-        let inv = Montgomery::neginv(&m);
-        let a = Montgomery::transform(n, &m);
-        MontgomeryBigint {
-            a,
-            minv: Rc::new((m, inv)),
-        }
-    }
-}
-
 impl<T: Integer + Montgomery> PartialEq for MontgomeryInt<T> {
     #[inline]
     fn eq(&self, other: &Self) -> bool {
@@ -284,14 +299,6 @@ impl<T: Integer + Montgomery> PartialEq for MontgomeryInt<T> {
     }
 }
 
-impl<T: Integer + Montgomery> PartialEq for MontgomeryBigint<T> {
-    #[inline]
-    fn eq(&self, other: &Self) -> bool {
-        self.check_modulus_eq(other);
-        self.a == other.a
-    }
-}
-
 impl<T: Integer + Montgomery> Add for MontgomeryInt<T> {
     type Output = Self;
 
@@ -307,18 +314,6 @@ impl<T: Integer + Montgomery> Add for MontgomeryInt<T> {
     }
 }
 
-impl<T: Integer + Montgomery> Add for MontgomeryBigint<T> {
-    type Output = Self;
-
-    #[inline]
-    fn add(self, rhs: Self) -> Self::Output {
-        self.check_modulus_eq(&rhs);
-        let m = &Borrow::<(T, T::Inv)>::borrow(&self.minv).0;
-        let a = Montgomery::add(&self.a, &rhs.a, m);
-        MontgomeryBigint { a, minv: self.minv }
-    }
-}
-
 impl<T: Integer + Montgomery> Sub for MontgomeryInt<T> {
     type Output = Self;
 
@@ -334,18 +329,6 @@ impl<T: Integer + Montgomery> Sub for MontgomeryInt<T> {
     }
 }
 
-impl<T: Integer + Montgomery> Sub for MontgomeryBigint<T> {
-    type Output = Self;
-
-    #[inline]
-    fn sub(self, rhs: Self) -> Self::Output {
-        self.check_modulus_eq(&rhs);
-        let m = &Borrow::<(T, T::Inv)>::borrow(&self.minv).0;
-        let a = Montgomery::sub(&self.a, &rhs.a, m);
-        MontgomeryBigint { a, minv: self.minv }
-    }
-}
-
 impl<T: Integer + Montgomery> Neg for MontgomeryInt<T> {
     type Output = Self;
 
@@ -360,17 +343,6 @@ impl<T: Integer + Montgomery> Neg for MontgomeryInt<T> {
     }
 }
 
-impl<T: Integer + Montgomery> Neg for MontgomeryBigint<T> {
-    type Output = Self;
-
-    #[inline]
-    fn neg(self) -> Self::Output {
-        let m = &Borrow::<(T, T::Inv)>::borrow(&self.minv).0;
-        let a = Montgomery::neg(&self.a, m);
-        MontgomeryBigint { a, minv: self.minv }
-    }
-}
-
 impl<T: Integer + Montgomery> Mul for MontgomeryInt<T> {
     type Output = Self;
 
@@ -386,18 +358,6 @@ impl<T: Integer + Montgomery> Mul for MontgomeryInt<T> {
     }
 }
 
-impl<T: Integer + Montgomery> Mul for MontgomeryBigint<T> {
-    type Output = Self;
-
-    #[inline]
-    fn mul(self, rhs: Self) -> Self::Output {
-        self.check_modulus_eq(&rhs);
-        let minv = Borrow::<(T, T::Inv)>::borrow(&self.minv);
-        let a = Montgomery::mul(&self.a, &rhs.a, &minv.0, &minv.1);
-        MontgomeryBigint { a, minv: self.minv }
-    }
-}
-
 impl<T: Integer + Montgomery> Pow<T> for MontgomeryInt<T> {
     type Output = Self;
 
@@ -412,45 +372,6 @@ impl<T: Integer + Montgomery> Pow<T> for MontgomeryInt<T> {
     }
 }
 
-impl<T: Integer + Montgomery> Pow<T> for MontgomeryBigint<T> {
-    type Output = Self;
-
-    #[inline]
-    fn pow(self, rhs: T) -> Self::Output {
-        let minv = Borrow::<(T, T::Inv)>::borrow(&self.minv);
-        let a = Montgomery::pow(&self.a, &rhs, &minv.0, &minv.1);
-        MontgomeryBigint { a, minv: self.minv }
-    }
-}
-
-impl<T: Integer + Montgomery + Clone> ModularInteger for MontgomeryBigint<T>
-where
-    T::Double: From<T>,
-{
-    type Base = T;
-
-    #[inline]
-    fn modulus(&self) -> &T {
-        &Borrow::<(T, T::Inv)>::borrow(&self.minv).0
-    }
-
-    #[inline]
-    fn residue(&self) -> T {
-        let minv = Borrow::<(T, T::Inv)>::borrow(&self.minv);
-        Montgomery::reduce(T::Double::from(self.a.clone()), &minv.0, &minv.1)
-    }
-
-    #[inline]
-    fn new(&self, n: T) -> Self {
-        let m = &Borrow::<(T, T::Inv)>::borrow(&self.minv).0;
-        let a = Montgomery::transform(n, &m);
-        MontgomeryBigint {
-            a,
-            minv: self.minv.clone(),
-        }
-    }
-}
-
 impl<T: Integer + Montgomery + Clone> ModularInteger for MontgomeryInt<T>
 where
     T::Double: From<T>,
