@@ -1,6 +1,10 @@
 use crate::{ModularCoreOps, ModularOps};
 use num_integer::Integer;
 
+// TODO: implement the modular functions as const: https://github.com/rust-lang/rust/pull/68847
+// TODO: provide utility functions to convert signed integers to unsigned during modular operation
+// (especially negm and absm)
+
 macro_rules! impl_powm_uprim {
     ($T:ty) => {
         fn powm(self, exp: $T, m: &$T) -> $T {
@@ -59,7 +63,7 @@ macro_rules! impl_jacobi_uprim {
                         t *= -1;
                     }
                 }
-                std::mem::swap(&mut a, &mut n);
+                core::mem::swap(&mut a, &mut n);
                 if (a & 3) == 3 && (n & 3) == 3 {
                     t *= -1;
                 }
@@ -207,6 +211,7 @@ impl ModularCoreOps<u128, &u128> for u128 {
     }
 
     // TODO: benchmark against http://www.janfeitsma.nl/math/psp2/expmod
+    // TODO: benchmark against udouble implementation
     fn mulm(self, rhs: u128, m: &u128) -> u128 {
         if let Some(ab) = self.checked_mul(rhs) {
             return ab % m;
